from __future__ import unicode_literals

import datetime
import threading

from . import album
from . import artist
from . import dispatch
from . import listener
from . import request
from . import schedule
from . import song

pre_sync = dispatch.Signal()
post_sync = dispatch.Signal()


class RainwaveChannel(dict):
    """A :class:`RainwaveChannel` object represents one channel on the Rainwave
    network.

    .. note::

        You should not instantiate an object of this class directly, but rather
        obtain one from :attr:`RainwaveClient.channels`.
    """

    def __init__(self, client, raw_info):
        self._client = client
        super(RainwaveChannel, self).__init__(raw_info)
        self._do_sync = False
        self._sync_thread = None

        self._raw_albums = None
        self._albums = None
        self._raw_artists = None
        self._artists = None

        self._sched_current = {}
        self._sched_next = []
        self._sched_history = []
        self._sched_lock = threading.Lock()

        self._raw_requests = []
        self._raw_user_requests = []
        self._requests_lock = threading.Lock()

    def __repr__(self):
        return '<RainwaveChannel [{0}]>'.format(self.name)

    def __str__(self):
        return '{0}: {1}'.format(self.name, self.description)

    def _do_async_get(self):
        if not self._stale():
            return
        d = self.client.call('/info', {'sid': self.id})
        with self._sched_lock:
            self._sched_current = d['sched_current']
            self._sched_next = d['sched_next']
            self._sched_history = d['sched_history']
        with self._requests_lock:
            self._raw_requests = d['request_line']
            self._raw_user_requests = d['requests']
        post_sync.send(self)

    def _do_sync_thread(self):
        self._do_sync = True
        while self._do_sync:
            pre_sync.send(self)
            args = {'sid': self.id}
            if not self._sched_current:
                args['resync'] = 'true'
            d = self.client.call('sync', args)
            if self._do_sync:
                with self._sched_lock:
                    self._sched_current = d['sched_current']
                    self._sched_next = d['sched_next']
                    self._sched_history = d['sched_history']
                with self._requests_lock:
                    self._raw_requests = d['request_line']
                    self._raw_user_requests = d['requests']
                post_sync.send(self, channel=self)

    def _get_listener_raw_info(self, listener_id):
        args = {'id': listener_id, 'sid': self.id}
        d = self.client.call('listener', args)
        if 'listener' in d:
            return d['listener']
        err = 'There is no listener with id: {0}'.format(listener_id)
        raise IndexError(err)

    def _new_schedule(self, raw_schedule):
        if raw_schedule['type'] == 'Election':
            return schedule.RainwaveElection(self, raw_schedule)
        if raw_schedule['type'] == 'OneUp':
            return schedule.RainwaveOneTimePlay(self, raw_schedule)

    def _stale(self):
        """Return True if timeline information (:attr:`schedule_current`,
        :attr:`schedule_next`, and :attr:`schedule_history`) is missing or out
        of date."""

        if len(self._sched_next) < 1:
            return True
        now = datetime.datetime.utcnow()
        ts = self._sched_current['end']
        ts = datetime.datetime.utcfromtimestamp(ts)
        return now > ts

    @property
    def albums(self):
        """A list of :class:`RainwaveAlbum` objects in the playlist of the
        channel."""

        if self._raw_albums is None:
            d = self.client.call('all_albums', {'sid': self.id})
            if 'all_albums' in d:
                self._raw_albums = d['all_albums']
        if self._albums is None:
            self._albums = []
            for raw_album in self._raw_albums:
                new_album = album.RainwaveAlbum(self, raw_album)
                self._albums.append(new_album)
        return self._albums

    @property
    def artists(self):
        """A list of :class:`RainwaveArtist` objects in the playlist of the
        channel."""

        if self._raw_artists is None:
            d = self.client.call('all_artists', {'sid': self.id})
            if 'all_artists' in d:
                self._raw_artists = d['all_artists']
        if self._artists is None:
            self._artists = []
            for raw_artist in self._raw_artists:
                new_artist = artist.RainwaveArtist(self, raw_artist)
                self._artists.append(new_artist)
        return self._artists

    def clear_rating(self, song_id):
        args = {'sid': self.id, 'song_id': song_id}
        return self.client.call('clear_rating', args)

    @property
    def client(self):
        """The :class:`RainwaveClient` object that the channel belongs to."""
        return self._client

    def delete_request(self, song_id):
        args = {'song_id': song_id, 'sid': self.id}
        d = self.client.call('delete_request', args)
        if d['delete_request_result']['success']:
            with self._requests_lock:
                self._raw_user_requests = d['requests']
            return d
        else:
            raise Exception(d['delete_request_result']['text'])

    @property
    def description(self):
        """A description of the channel."""
        return self['description']

    def fave_album(self, album_id, fave):
        args = {'album_id': album_id, 'fave': fave, 'sid': self.id}
        return self.client.call('fave_album', args)

    def fave_song(self, song_id, fave):
        args = {'song_id': song_id, 'fave': fave}
        return self.client.call('fave_song', args)

    def get_album_by_id(self, album_id):
        """Return a :class:`RainwaveAlbum` for the given album ID. Raise an
        :exc:`IndexError` if there is no album with the given ID in the
        playlist of the channel.

        :param album_id: the ID of the desired album.
        :type album_id: int
        """

        for alb in self.albums:
            if alb.id == album_id:
                args = {'sid': self.id, 'id': album_id}
                d = self.client.call('album', args)
                return album.RainwaveAlbum(self, d['album'])
        error = 'Channel does not contain album with id: {0}'.format(album_id)
        raise IndexError(error)

    def get_album_by_name(self, name):
        """Return a :class:`RainwaveAlbum` for the given album name. Raise an
        :exc:`IndexError` if there is no album with the given name in the
        playlist of the channel.

        :param name: the name of the desired album.
        :type name: str
        """

        for alb in self.albums:
            if alb.name == name:
                return alb
        error = 'Channel does not contain album with name: {0}'.format(name)
        raise IndexError(error)

    def get_artist_by_id(self, artist_id):
        """Return a :class:`RainwaveArtist` for the given artist ID. Raise an
        :exc:`IndexError` if there is no artist with the given ID in the
        playlist of the channel.

        :param artist_id: the ID of the desired artist.
        :type artist_id: int
        """

        args = {'sid': self.id, 'id': artist_id}
        d = self.client.call('artist', args)
        if 'id' in d['artist']:
            return artist.RainwaveArtist(self, d['artist'])
        err = 'Channel does not contain artist with id: {0}'.format(artist_id)
        raise IndexError(err)

    def get_listener_by_id(self, listener_id):
        """Return a :class:`RainwaveListener` for the given listener ID. Raise
        an :exc:`IndexError` if there is no listener with the given ID.

        :param listener_id: the ID of the desired listener.
        :type listener_id: int
        """

        raw_listener = self._get_listener_raw_info(listener_id)
        return listener.RainwaveListener(self, raw_listener)

    def get_listener_by_name(self, name):
        """Return a :class:`RainwaveListener` for the given listener name. Raise
        an :exc:`IndexError` if there is no listener with the given name
        currently listening to the channel.

        :param name: the name of the desired listener.
        :type name: str
        """

        for _listener in self.listeners:
            if _listener.name == name:
                return _listener
        err = 'No current listener named {0}'.format(name)
        raise IndexError(err)

    def get_song_by_id(self, song_id):
        """Return a :class:`RainwaveSong` for the given song ID. Raise an
        :exc:`IndexError` if there is no song with the given ID in the playlist
        of the channel.

        :param song_id: the ID of the desired song.
        :type song_id: int
        """

        args = {'sid': self.id, 'id': song_id}
        d = self.client.call('song', args)
        if 'albums' in d['song']:
            alb = self.get_album_by_id(d['song']['albums'][0]['id'])
            return song.RainwaveSong(alb, d['song'])
        err = 'Channel does not contain song with id: {0}'.format(song_id)
        raise IndexError(err)

    @property
    def id(self):
        """The ID of the channel."""
        return self['id']

    @property
    def listeners(self):
        """A list of :class:`RainwaveListener` objects listening to the
        channel."""
        _listeners = []
        d = self.client.call('current_listeners', {'sid': self.id})
        for raw_listener in d['current_listeners']:
            _listeners.append(listener.RainwaveListener(self, raw_listener))
        return _listeners

    @property
    def name(self):
        """The name of the channel."""
        return self['name']

    @property
    def ogg_stream(self):
        """The URL of the OGG stream for the channel. See also
        :attr:`mp3_stream`."""
        return self.mp3_stream.replace('.mp3', '.ogg')

    def rate(self, song_id, rating):
        args = {'sid': self.id, 'song_id': song_id, 'rating': rating}
        return self.client.call('rate', args)

    def reorder_requests(self, order):
        args = {'sid': self.id, 'order': order}
        d = self.client.call('order_requests', args)
        if d['order_requests_result']['success']:
            with self._requests_lock:
                self._raw_user_requests = d['requests']
            return d
        else:
            raise Exception(d['order_requests_result']['text'])

    def request_song(self, song_id):
        args = {'song_id': song_id, 'sid': self.id}
        d = self.client.call('request', args)
        if d['request_result']['success']:
            with self._requests_lock:
                self._raw_user_requests = d['requests']
            return d
        else:
            raise Exception(d['request_result']['text'])

    def clear_requests(self):
        return self.client.call('clear_requests', {'sid': self.id})

    @property
    def requests(self):
        """A list of :class:`RainwaveRequest` objects in the request line of
        the channel."""
        if self._stale():
            self._do_async_get()
        rqs = []
        with self._requests_lock:
            for raw_request in self._raw_requests:
                if raw_request.get('song_id') is None:
                    continue
                _song = self.get_song_by_id(raw_request['song_id'])
                _reqr = self.get_listener_by_id(raw_request['user_id'])
                rq = request.RainwaveRequest.request_from_song(_song, _reqr)
                rqs.append(rq)
        return rqs

    @property
    def schedule_current(self):
        """The current :class:`RainwaveSchedule` for the channel."""
        if self._stale():
            self._do_async_get()
        with self._sched_lock:
            sched_current = self._new_schedule(self._sched_current)
        return sched_current

    @property
    def schedule_history(self):
        """A list of the past :class:`RainwaveSchedule` objects for the
        channel."""
        if self._stale():
            self._do_async_get()
        sched_history = []
        with self._sched_lock:
            for raw_sched in self._sched_history:
                sched_history.append(self._new_schedule(raw_sched))
        return sched_history

    @property
    def schedule_next(self):
        """A list of the next :class:`RainwaveSchedule` objects for the
        channel."""
        if self._stale():
            self._do_async_get()
        sched_next = []
        with self._sched_lock:
            for raw_sched in self._sched_next:
                sched_next.append(self._new_schedule(raw_sched))
        return sched_next

    def start_sync(self):
        """Begin syncing the timeline for the channel."""

        self.stop_sync()
        self._sync_thread = threading.Thread(target=self._do_sync_thread)
        self._sync_thread.daemon = True
        self._sync_thread.start()

    def stop_sync(self):
        """Stop syncing the timeline for the channel."""

        self._do_sync = False
        self._sync_thread = None

    @property
    def mp3_stream(self):
        """The URL of the MP3 stream for the channel. See also
        :attr:`ogg_stream`."""
        return self['stream']

    @property
    def user_requests(self):
        """A :class:`RainwaveUserRequestQueue` of :class:`RainwaveUserRequest`
        objects in the listener's personal request queue."""
        if self._stale():
            self._do_async_get()
        rqs = request.RainwaveUserRequestQueue(self)
        with self._requests_lock:
            for raw_request in self._raw_user_requests:
                album_id = raw_request['albums'][0]['id']
                alb = self.get_album_by_id(album_id)
                rq = request.RainwaveUserRequest(alb, raw_request)
                rqs.append(rq)
        return rqs

    def vote(self, entry_id):
        args = {'entry_id': entry_id, 'sid': self.id}
        return self.client.call('vote', args)



from __future__ import unicode_literals

import datetime

from . import category


class RainwaveAlbum(dict):
    """A :class:`RainwaveAlbum` object represents one album.

    .. note::

        You should not instantiate an object of this class directly, but rather
        obtain one from :attr:`RainwaveChannel.albums`.
    """

    def __init__(self, channel, raw_info):
        self._channel = channel
        super(RainwaveAlbum, self).__init__(raw_info)

    def __repr__(self):
        _repr = '<RainwaveAlbum [{0} // {1}]>'
        return _repr.format(self.channel.name, self.name)

    def __str__(self):
        _str = '{0} // {1}'
        return _str.format(self.channel.name, self.name)

    @property
    def art(self):
        """The URL of the cover art for the album."""
        return self.channel.client.art_fmt.format(self['art'])

    @property
    def added_on(self):
        """A :class:`datetime.datetime` object specifying when the album was
        added to the playlist."""
        return datetime.datetime.utcfromtimestamp(self['added_on'])

    @property
    def categories(self):
        """A list of :class:`RainwaveCategory` objects representing the
        categories the songs in the album belong to."""
        if 'category_objects' not in self:
            self['category_objects'] = []
            for raw_cat in self['genres']:
                category_id = raw_cat['id']
                name = raw_cat['name']
                cat = category.RainwaveCategory(self.channel, category_id, name)
                self['category_objects'].append(cat)
        return self['category_objects']

    @property
    def channel(self):
        """The :class:`RainwaveChannel` object the album belongs to."""
        return self._channel

    @property
    def cool(self):
        """A boolean representing whether the entire album is on cooldown.
        :attr:`cool` will be `True` if and only if every song in the album is on
        cooldown."""
        return self['cool']

    @property
    def cool_lowest(self):
        """A :class:`datetime.datetime` object specifying the earliest date and
        time a song in the album will be out of cooldown and available to play.
        If any song in the album is already available, :attr:`cool_lowest` will
        be in the past."""
        return datetime.datetime.utcfromtimestamp(self['cool_lowest'])

    @property
    def fave(self):
        """A boolean representing whether the album is marked as a fave or not.
        Change whether the album is a fave by assigning a boolean value to this
        attribute."""
        return self.get('fave', False)

    @fave.setter
    def fave(self, value):
        value = bool(value)
        if value == self.fave:
            return
        d = self.channel.fave_album(self.id, str(value).lower())
        if d['fave_album_result']['success']:
            self['fave'] = value
        else:
            raise Exception(d['fave_album_result']['text'])

    @property
    def fave_count(self):
        """The number of listeners who have marked the album as a favourite."""
        return self['fave_count']

    @property
    def id(self):
        """The ID of the album."""
        return self['id']

    @property
    def name(self):
        """The name of the album."""
        return self['name']

    @property
    def played_last(self):
        """A :class:`datetime.datetime` object specifying the most recent date
        and time when a song in the album played."""
        return datetime.datetime.utcfromtimestamp(self['played_last'])

    @property
    def rating(self):
        """The average of all ratings given to songs in the album by only the
        listener authenticating to the API."""
        return self['rating_user']

    @property
    def rating_avg(self):
        """The average of all ratings given to songs in the album by all
        listeners."""
        return self['rating']

    @property
    def rating_complete(self):
        """A boolean representing whether the listener has rated all songs in
        the album."""
        return self['rating_complete']

    @property
    def rating_count(self):
        """The total number of ratings given to songs in the album by all
        listeners."""
        return self['rating_count']

    @property
    def rating_histogram(self):
        """A dictionary representing the distribution of ratings given to all
        songs in the album by all listeners. For example::

            >>> album.rating_histogram
            {'1.0': 4, '1.5': 4, '2.0': 6, ..., '4.5': 46, '5.0': 26}
        """
        return self['rating_histogram']

    @property
    def rating_rank(self):
        """The position of the album when albums on the channel are ranked by
        rating. The highest-rated album will have :attr:`rating_rank` == 1."""
        return self['rating_rank']

    @property
    def rating_user(self):
        """See :attr:`rating`."""
        return self.rating

    @property
    def request_count(self):
        """The total number of times a song in the album was requested by any
        listener."""
        return self['request_count']

    @property
    def request_rank(self):
        """The position of the album when albums on the channel are ranked by
        how often they are requested. The most-requested album will have
        :attr:`request_rank` == 1."""
        return self['request_rank']

    @property
    def songs(self):
        """A list of :class:`RainwaveSong` objects in the album."""
        if 'song_objects' not in self:
            self['song_objects'] = []
            if 'songs' not in self:
                album_obj = self.channel.get_album_by_id(self.id)
                self['songs'] = album_obj['songs']
            for raw_song in self['songs']:
                new_song = self.channel.get_song_by_id(raw_song['id'])
                self['song_objects'].append(new_song)
        return self['song_objects']

    @property
    def vote_count(self):
        """The total number of election votes songs in the album have
        received."""
        return self['vote_count']

    def get_song_by_id(self, song_id):
        """ Return a :class:`RainwaveSong` for the given song ID. Raises an
        :exc:`IndexError` if there is no song with the given ID in the
        album.

        :param song_id: the ID of the desired song.
        :type song_id: int
        """

        for song in self.songs:
            if song.id == song_id:
                return song
        err = 'Album does not contain song with id: {0}'.format(song_id)
        raise IndexError(err)



"""
Signalling framework for internal use.

This is loosely inspired by the Django signalling systems, although far more
minimalistic.

To use, either register a callable using your target signal's
:method:`Signal.connect(receiver)` or use :decorator:`receiver` like so:

    from rainwaveclient.channel import post_sync
    from rainwaveclient.dispatch import receiver

    @receiver(post_sync)
    def on_post_sync(signal, sender, **kwargs):
        ## do something here
"""

from __future__ import unicode_literals

import threading


class Signal:
    """A signal is triggered each time a particular event happens."""

    def __init__(self):
        self.receivers = set()
        self.lock = threading.Lock()

    def connect(self, _receiver):
        """Add a receiver to the signal."""

        with self.lock:
            self.receivers.add(_receiver)

    def disconnect(self, _receiver):
        """Remove a receiver from the signal."""

        with self.lock:
            self.receivers.remove(_receiver)

    def send(self, sender, **kwargs):
        """Send the signal to all connected receivers."""

        for _receiver in self.receivers:
            _receiver(signal=self, sender=sender, **kwargs)


def receiver(signal, **kwargs):
    """Decorator for registering a signal."""

    def decorator(func):
        signal.connect(func, **kwargs)
    return decorator



from __future__ import unicode_literals


class RainwaveArtist(dict):
    """A :class:`RainwaveArtist` object represents one artist.

    .. note::

        You should not instantiate an object of this class directly, but rather
        obtain one from :attr:`RainwaveChannel.artists` or
        :attr:`RainwaveSong.artists`.
    """

    def __init__(self, channel, raw_info):
        self._channel = channel
        super(RainwaveArtist, self).__init__(raw_info)

    def __repr__(self):
        return '<RainwaveArtist [{0}]>'.format(self)

    def __str__(self):
        return self.name

    @property
    def channel(self):
        """The :class:`RainwaveChannel` object associated with the artist."""
        return self._channel

    @property
    def id(self):
        """The ID of the artist."""
        return self['id']

    @property
    def name(self):
        """The name of the artist."""
        return self['name']

    @property
    def song_count(self):
        """The number of songs attributed to the artist."""
        return len(self.songs)

    @property
    def songs(self):
        """A list of :class:`RainwaveSong` objects attributed to the artist."""
        if 'song_objects' not in self:
            self['song_objects'] = []
            for chan_id, albums in self['all_songs'].items():
                for album_id, album_songs in albums.items():
                    for raw_song in album_songs:
                        song = self.channel.get_song_by_id(raw_song['id'])
                        self['song_objects'].append(song)
        return self['song_objects']



from __future__ import unicode_literals


class RainwaveCategory:

    def __init__(self, channel, category_id, name):
        self._channel = channel
        self.category_id = category_id
        self.name = name



from __future__ import unicode_literals

import datetime

from . import song


class RainwaveSchedule(dict):
    """A :class:`RainwaveSchedule` object represents an event on a channel.

    .. note::

        You should not instantiate an object of this class directly, but rather
        obtain one from :attr:`RainwaveChannel.schedule_current`,
        :attr:`RainwaveChannel.schedule_next`, or
        :attr:`RainwaveChannel.schedule_history`.
    """

    def __init__(self, channel, raw_info):
        self._channel = channel
        super(RainwaveSchedule, self).__init__(raw_info)

    def __repr__(self):
        return '<RainwaveSchedule [{0}]>'.format(self.channel.name)

    def __str__(self):
        return repr(self)

    @property
    def channel(self):
        """The :class:`RainwaveChannel` object the event belongs to."""
        return self._channel

    @property
    def id(self):
        """The ID of the event."""
        return self['id']

    @property
    def start(self):
        """A `datetime.datetime` object representing the start time of the event
        in UTC time."""
        return datetime.datetime.utcfromtimestamp(self['start'])


class RainwaveElection(RainwaveSchedule):
    """A :class:`RainwaveElection` object is a subclass of
    :class:`RainwaveSchedule` and represents an election event on a channel."""

    def __repr__(self):
        return '<RainwaveElection [{0}]>'.format(self.channel.name)

    @property
    def candidates(self):
        """A list of :class:`RainwaveCandidate` objects in the election."""
        if 'candidate_objects' not in self:
            self['candidate_objects'] = []
            for raw_song in self['songs']:
                alb = self.channel.get_album_by_id(raw_song['albums'][0]['id'])
                tmp_song = song.RainwaveCandidate(alb, raw_song)
                self['candidate_objects'].append(tmp_song)
        return self['candidate_objects']

    @property
    def songs(self):
        """See :attr:`candidates`."""
        return self.candidates


class RainwaveOneTimePlay(RainwaveSchedule):
    """A :class:`RainwaveOneTimePlay` object is a subclass of
    :class:`RainwaveSchedule` and represents a song added directly to the
    timeline by a manager."""

    def __repr__(self):
        return '<RainwaveOneTimePlay [{0}]>'.format(self.channel.name)

    @property
    def name(self):
        """The name of the event."""
        return '{0} Power Hour'.format(self['name'])

    @property
    def song(self):
        """The :class:`RainwaveSong` for the event."""
        album_id = self['songs'][0]['albums'][0]['id']
        tmp_album = self.channel.get_album_by_id(album_id)
        return song.RainwaveSong(tmp_album, self['songs'][0])



from __future__ import unicode_literals

import json
import logging
import sys
import uuid

from . import channel

if sys.version_info[0] == 2:
    from urllib import urlencode
    from urllib2 import urlopen, HTTPError, Request
else:
    from urllib.error import HTTPError
    from urllib.request import urlopen, Request
    from urllib.parse import urlencode

log = logging.getLogger(__name__)


class RainwaveClient:
    """A :class:`RainwaveClient` object provides a simple interface to the
    Rainwave API (see https://rainwave.cc/api4/ for details about the API).

    :param user_id: the User ID to use when communicating with the API.
    :param key: the API key to use when communicating with the API.
    """

    #: The URL upon which all API calls are based.
    base_url = 'https://rainwave.cc/api4/'

    #: The format string used to build canonical album art URLs.
    art_fmt = 'https://rainwave.cc{0}_320.jpg'

    def __init__(self, user_id=None, key=None):
        if user_id is not None:
            self._user_id = int(user_id)
        if key is not None:
            self._key = key
        self._raw_channels = None
        self._channels = None
        self.user_agent = uuid.uuid4().hex

    def __repr__(self):
        msg = 'RainwaveClient(user_id={0!r}, key={1!r})'
        return msg.format(self.user_id, self.key)

    def call(self, path, args=None):
        """Make a direct call to the API if you know the necessary path and
        arguments.

        :param path: the URL path of the API method to call.
        :type path: str
        :param args: (optional) any arguments required by the API method.
        :type args: dict

        Usage::

          >>> from rainwaveclient import RainwaveClient
          >>> rw = RainwaveClient(5049, 'abcde12345')
          >>> rw.call('album', {'id': 1, 'sid': 5})
          {'album': {'name': 'Bravely Default: Flying Fairy', ...}}
        """

        url = '{0}{1}'.format(self.base_url, path.lstrip('/'))

        if args is None:
            args = {}
        if 'user_id' not in args and self.user_id:
            args['user_id'] = self.user_id
        if 'key' not in args and self.key:
            args['key'] = self.key

        data = urlencode(args).encode()
        headers = {'user-agent': self.user_agent}
        req = Request(url=url, data=data, headers=headers)
        try:
            response = urlopen(req)
        except HTTPError as e:
            response = e
        body = response.read().decode(encoding='utf-8')
        api_response = json.loads(body)
        log.debug(api_response)
        return api_response

    @property
    def channels(self):
        """A list of :class:`RainwaveChannel` objects associated with this
        :class:`RainwaveClient` object."""

        if self._raw_channels is None:
            d = self.call('stations')
            if 'stations' in d:
                self._raw_channels = d['stations']
            else:
                raise Exception

        if self._channels is None:
            self._channels = list()
            for raw_channel in self._raw_channels:
                new_channel = channel.RainwaveChannel(self, raw_channel)
                self._channels.append(new_channel)

        return self._channels

    @property
    def key(self):
        """The API key to use when communicating with the API. Find your API
        key at http://rainwave.cc/keys/."""
        return self._key

    @key.setter
    def key(self, value):
        self._key = value

    @property
    def user_id(self):
        """The User ID to use when communicating with the API. Find your User ID
        at http://rainwave.cc/keys/."""
        return self._user_id

    @user_id.setter
    def user_id(self, value):
        self._user_id = value



from __future__ import unicode_literals


class RainwaveListener(dict):
    """A :class:`RainwaveListener` object represents a radio listener."""

    #: The :class:`RainwaveChannel` the listener belongs to.
    channel = None

    def __init__(self, channel, raw_info):
        self.channel = channel
        super(RainwaveListener, self).__init__(raw_info)

    def __repr__(self):
        return '<RainwaveListener [{0}]>'.format(self)

    def __str__(self):
        return self.name

    @property
    def avatar(self):
        """The URL of the listener's avatar."""
        return 'http://rainwave.cc{0}'.format(self['avatar'])

    @property
    def color(self):
        """See :attr:`colour`."""
        return self.colour

    @property
    def colour(self):
        """A hexadecimal string representing the listener's colour on the
        forums."""
        return self['colour']

    @property
    def id(self):
        """The ID of the listener."""
        if 'id' in self:
            return self['id']
        return self['user_id']

    @property
    def losing_requests(self):
        """The number of requests made by the listener that lost their
        election."""
        return self['losing_requests']

    @property
    def losing_votes(self):
        """The number of votes the listeners has given to a song that lost an
        election."""
        return self['losing_votes']

    @property
    def mind_changes(self):
        """The total number of times the listener changed a song rating."""
        return self['mind_changes']

    @property
    def name(self):
        """The name of the listener."""
        return self['name']

    @property
    def rank(self):
        """A string representing the listener's title on the forums."""
        return self['rank']

    @property
    def total_ratings(self):
        """The total number of songs the listener has rated."""
        return self['total_ratings']

    @property
    def total_requests(self):
        """The total number of requests the listener has made."""
        return self['total_requests']

    @property
    def total_votes(self):
        """The number of votes the listener has cast in the last two weeks."""
        return self['total_votes']

    @property
    def user_id(self):
        """See :attr:`id`."""
        return self.id

    @property
    def winning_requests(self):
        """The number of requests made by the listener that won their
        election."""
        return self['winning_requests']

    @property
    def winning_votes(self):
        """The number of votes the listener has given to a song that won an
        election."""
        return self['winning_votes']



from __future__ import unicode_literals

from . import category


class RainwaveSong(dict):
    """A :class:`RainwaveSong` object represents one song.

    .. note::

        You should not instantiate an object of this class directly, but rather obtain one from
        :attr:`RainwaveAlbum.songs`, :attr:`RainwaveArtist.songs`, or some other object.
    """

    def __init__(self, album, raw_info):
        self._album = album
        super(RainwaveSong, self).__init__(raw_info)

    def __len__(self):
        return self['length']

    def __repr__(self):
        return '<RainwaveSong [{0}]>'.format(self)

    def __str__(self):
        msg = '{0} // {1} // {2}'
        return msg.format(self.album, self.title, self.artist_string)

    @property
    def album(self):
        """The :class:`RainwaveAlbum` object the song belongs to."""
        return self._album

    @property
    def artist_string(self):
        """A single string with the names of all artists for the song."""
        return ', '.join([artist.name for artist in self.artists])

    @property
    def artists(self):
        """A list of :class:`RainwaveArtist` objects the song is attributed to."""
        if 'artist_objects' not in self:
            self['artist_objects'] = []
            if 'artists' not in self:
                song_obj = self.album.channel.get_song_by_id(self.id)
                self['artists'] = song_obj['artists']
            for raw_artist in self['artists']:
                artist_id = raw_artist['id']
                channel = self.album.channel
                new_artist = channel.get_artist_by_id(artist_id)
                self['artist_objects'].append(new_artist)
        return self['artist_objects']

    @property
    def available(self):
        """A boolean representing whether the song is available to play or not. Opposite of :attr:`cool`."""
        return not self.cool

    @property
    def categories(self):
        """A list of :class:`RainwaveCategory` objects representing the categories the song belongs to."""
        if 'category_objects' not in self:
            self['category_objects'] = []
            for raw_cat in self['groups']:
                chan = self.album.channel
                cat_id = raw_cat['id']
                name = raw_cat['name']
                cat = category.RainwaveCategory(chan, cat_id, name)
                self['category_objects'].append(cat)
        return self['category_objects']

    @property
    def channel_id(self):
        """The :attr:`RainwaveChannel.id` of the channel the song belongs to."""
        return self['sid']

    @property
    def cool(self):
        """A boolean representing whether the song is on cooldown. Opposite of :attr:`available`."""
        return self['cool']

    @property
    def fave(self):
        """A boolean representing whether the song is marked as a fave or not. Change whether the song is a fave by
        assigning a boolean value to this attribute."""
        return self['fave']

    @fave.setter
    def fave(self, value):
        value = bool(value)
        if value == self.fave:
            return
        d = self.album.channel.fave_song(self.id, str(value).lower())
        if d['fave_song_result']['success']:
            self['fave'] = value
        else:
            raise Exception(d['fave_song_result']['text'])

    @property
    def id(self):
        """The ID of the song."""
        return self['id']

    @property
    def length(self):
        """The length of the song in seconds. :class:`RainwaveSong` objects also support `len(song)`."""
        return len(self)

    @property
    def link_text(self):
        """The link text that corresponds with :attr:`url`."""
        return self['link_text']

    @property
    def origin_channel_id(self):
        """The :attr:`RainwaveChannel.id` of the home channel for the song. This could be different from
        :attr:`channel_id` if the song is in the playlist of multiple channels."""
        return self['origin_sid']

    @property
    def origin_sid(self):
        """See :attr:`origin_channel_id`."""
        return self.origin_channel_id

    @property
    def rating(self):
        """The rating given to the song by the listener authenticating to the API. Change the rating by assigning a new
        value to this attribute."""
        return self['rating_user']

    @rating.setter
    def rating(self, value):
        if self.rating == value:
            return
        d = self.album.channel.rate(self.id, value)
        if d['rate_result']['success']:
            self['rating_user'] = value
        else:
            raise Exception(d['rate_result']['text'])

    @rating.deleter
    def rating(self):
        d = self.album.channel.clear_rating(self.id)
        if d['rate_result']['success']:
            self['rating_user'] = None
        else:
            raise Exception(d['rate_result']['text'])

    @property
    def rating_allowed(self):
        """A boolean representing whether the listener can currently rate the song."""
        return self['rating_allowed']

    @property
    def rating_avg(self):
        """The average of all ratings given to the song by all listeners."""
        return self['rating']

    @property
    def rating_count(self):
        """The total number of ratings given to the song by all listeners."""
        return self['rating_count']

    @property
    def rating_histogram(self):
        """A dictionary representing the distribution of ratings given to the song by all listeners.
        For example::

            >>> song.rating_histogram
            {'1.0': 4, '1.5': 4, '2.0': 6, ..., '4.5': 46, '5.0': 26}
        """
        return self['rating_histogram']

    @property
    def rating_rank(self):
        """The position of the album when albums on the channel are ranked by rating. The highest-rated album will have
        :attr:`rating_rank` == 1."""
        return self['rating_rank']

    @property
    def rating_user(self):
        """See :attr:`rating`."""
        return self.rating

    @property
    def request_count(self):
        """The total number of times the song has been requested by any listener."""
        return self['request_count']

    @property
    def request_rank(self):
        """The position of the song when songs on the channel are ranked by how often they are requested. The
        most-requested song will have :attr:`rating_rank` == 1."""
        return self['rating_rank']

    @property
    def sid(self):
        """See :attr:`channel_id`."""
        return self.channel_id

    @property
    def title(self):
        """The title of the song."""
        return self['title']

    @property
    def url(self):
        """The URL of more information about the song."""
        return self['url']

    def request(self):
        """Add the song to the authenticating listener's request queue."""
        self.album.channel.request_song(self.id)


class RainwaveCandidate(RainwaveSong):
    """A :class:`RainwaveCandidate` object is a subclass of :class:`RainwaveSong` representing a song that is a
    candidate in an election."""

    def __init__(self, album, raw_info):
        super(RainwaveCandidate, self).__init__(album, raw_info)

    def __repr__(self):
        return '<RainwaveCandidate [{0}]>'.format(self)

    @property
    def entry_id(self):
        """The election entry ID for this candidate. Used for voting."""
        return self['entry_id']

    @property
    def is_request(self):
        """A boolean representing whether the candidate is a listener request or not."""
        return self['elec_request_user_id'] > 0

    @property
    def requested_by(self):
        """The :class:`RainwaveListener` who requested the candidate, if the candidate is a request. ``None``
        otherwise."""
        if self.is_request:
            user_id = self['elec_request_user_id']
            return self.album.channel.get_listener_by_id(user_id)
        return None

    def vote(self):
        """Cast a vote for the candidate."""
        d = self.album.channel.vote(self.entry_id)
        if not d['vote_result']['success']:
            raise Exception(d['vote_result']['text'])

    @property
    def votes(self):
        """The number of votes this candidate received in the election. If the election has not ended, :attr:`votes`
        will be ``0``."""
        return self['entry_votes']



from __future__ import unicode_literals

import setuptools

import rainwaveclient

setuptools.setup(
    name=rainwaveclient.__setup_name__,
    version=rainwaveclient.__version__,
    author=rainwaveclient.__author__,
    author_email=rainwaveclient.__author_email__,
    url=rainwaveclient.__url__,
    description=rainwaveclient.__description__,
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=['rainwaveclient'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries'
    ],
    license='MIT License',
    keywords=rainwaveclient.__keywords__
)



from __future__ import unicode_literals

from . import song


class RainwaveRequest(song.RainwaveSong):
    """A :class:`RainwaveRequest` object is a subclass of :class:`RainwaveSong` representing a song that has been
    requested to play on the radio."""

    @classmethod
    def request_from_song(cls, _song, requester):
        request = cls(_song.album, dict(_song.items()))
        request['requester'] = requester
        return request

    def __repr__(self):
        return '<RainwaveRequest [{0}]>'.format(self)

    @property
    def requester(self):
        """The :class:`RainwaveListener` who made the request."""
        return self['requester']


class RainwaveUserRequest(song.RainwaveSong):
    """A :class:`RainwaveUserRequest` object is a subclass of :class:`RainwaveSong` representing a song in the
    authenticating listener's requests queue."""

    def __repr__(self):
        return '<RainwaveUserRequest [{0}]>'.format(self)

    @property
    def blocked(self):
        """``True`` if the request is currently blocked. See :attr:`blocked_by_album` and :attr:`blocked_by_category` to
        determine why the request is blocked."""
        return bool(self.blocked_by_album or self.blocked_by_category)

    @property
    def blocked_by_album(self):
        """``True`` if the request is currently blocked because a song from the same album is in an election."""
        return self['elec_blocked_by'] == 'album'

    @property
    def blocked_by_category(self):
        """``True`` if the request is currently blocked because a song from the same category is in an election."""
        return self['elec_blocked_by'] == 'group'

    def delete(self):
        """Remove the requested song from the authenticating listener's request queue."""
        self.album.channel.delete_request(self.id)


class RainwaveUserRequestQueue(list):
    """A :class:`RainwaveUserRequestQueue` is a list-like object that supports explicit reordering."""

    def __init__(self, channel):
        self._channel = channel
        super(RainwaveUserRequestQueue, self).__init__()

    def clear(self):
        """Clear all requests from the queue."""
        d = self._channel.clear_requests()

    def reorder(self, order):
        """Change the order of the requests in the queue.

        :param order: the indices of the requests in the new order.
        :type order: sequence

        Example usage:

        If you have four songs in your request queue and you want to move the last song to the top of the queue::

        >>> from rainwaveclient import RainwaveClient
        >>> rw = RainwaveClient(5049, 'abcde12345')
        >>> game = rw.channels[0]
        >>> rq = game.user_requests
        >>> rq.reorder([3, 0, 1, 2])

        To randomly shuffle your request queue::

        >>> import random
        >>> indices = range(len(game.user_requests))
        >>> random.shuffle(indices)
        >>> rq.reorder(indices)

        All indices must appear in ``order`` and each index must only appear once.
        """

        if set(order) != set(range(len(self))):
            raise Exception('Incorrect indices.')
        if len(order) != len(set(order)):
            raise Exception('Wrong number of indices.')
        song_ids = ','.join([str(self[i].id) for i in order])
        self._channel.reorder_requests(song_ids)



from __future__ import unicode_literals

__setup_name__ = 'python-rainwave-client'
__version__ = '0.7.0'
__author__ = 'William Jackson'
__author_email__ = 'william@subtlecoolness.com'
__url__ = 'https://github.com/williamjacksn/python-rainwave-client'
__description__ = 'Python client library for Rainwave'
__keywords__ = 'rainwave client'

from .client import RainwaveClient
from .channel import RainwaveChannel
from .album import RainwaveAlbum
from .artist import RainwaveArtist
from .category import RainwaveCategory
from .schedule import RainwaveSchedule, RainwaveElection, RainwaveOneTimePlay
from .song import RainwaveSong, RainwaveCandidate
from .request import RainwaveRequest, RainwaveUserRequest
from .request import RainwaveUserRequestQueue
from .listener import RainwaveListener
